﻿// ゲームとしてのメイン部分

using DxLibDLL;
using System;

namespace Tetris
{
    /// <summary>
    /// ゲームとしてのメイン部分
    /// </summary>
    class GameMain
    {
        //------------------------------------------------------------
        //  画面に配置するオブジェクト
        //------------------------------------------------------------

        // 背景
        Background background = new Background();

        // フィールドデータ
        Field fieldData = new Field();

        // メッセージ
        Message message = new Message();

        // プレイヤー（操作中の）ブロック
        PlayerPiece playerPiece = new PlayerPiece();

        //------------------------------------------------------------
        //  制御用のクラス
        //------------------------------------------------------------

        // キーボード押下状態
        InputController inputController = new InputController();

        // Nextピース制御用
        NextPiece nextPiece = new NextPiece();

        //------------------------------------------------------------
        //  ゲームの状態を表す定義
        //------------------------------------------------------------

        // プレイヤーの状態（ゲームの状態とも言える）
        enum PlayerStatus
        {
            // 入力待ち
            InputWait,
            // 終了
            End,
        }
        PlayerStatus playerStatus;

        // プレイヤー操作ブロックの状態
        PlayerPieceStatus playerPieceStatus = new PlayerPieceStatus();

        /// <summary>
        /// ゲームとしてのメイン関数
        /// </summary>
        public void Main()
        {
            //------------------------------------------------------------
            //  イニシャル
            //------------------------------------------------------------

            // イニシャル処理。失敗したら終了
            if (Init() == false) return;

            //------------------------------------------------------------
            //  ゲームループ
            //------------------------------------------------------------

            // FPS制御用
            FPSTimer t = new FPSTimer(GL.FPS);

            // ゲームループ
            while (true)
            {
                // デバッグ用の緊急避難（ESCキー入力で強制終了）
                if (DX.CheckHitKey(DX.KEY_INPUT_ESCAPE) != 0) break;

                // 次のフレームを待つ
                t.WaitNextFrame();

                // 演算処理
                Perform();

                // 更新処理
                Update();

                // 描画可能の場合は描画処理
                if (t.IsDraw())
                {
                    // 描画先を裏画面に設定する
                    DX.SetDrawScreen(DX.DX_SCREEN_BACK);

                    // 画面に描かれているものを一回全部消す
                    DX.ClsDrawScreen();

                    // ここで描画処理をさせる
                    Draw();

#if DEBUG
                    // FPSを表示
                    DX.DrawString(0, 0, t.GetFpsReal().ToString("d3") + "/" + t.GetFps().ToString("d3"), 0x7F7F7F);
#endif

                    // 裏画面の内容を面画面に反映される
                    DX.ScreenFlip();

                    // 描画したときだけFPS計測させる
                    t.CalcFps();
                }

                // Windowメッセージを処理させる
                if (DX.ProcessMessage() == -1)
                {
                    break;
                }
            }

            //------------------------------------------------------------
            //  終了処理
            //------------------------------------------------------------

            // 終了処理
            Final();
        }

        /// <summary>
        /// イニシャル処理
        /// </summary>
        /// <returns>true:正常 false:異常</returns>
        private bool Init()
        {
            // プレイヤーの状態をセット
            playerStatus = PlayerStatus.InputWait;

            // プレイヤー操作ブロックの状態をリセット
            playerPieceStatus.ResetCount();

            // 初期メッセージ
            message.Set("");    // 何も表示しない

            // キーボードリピート状態の開始カウンタ値のセット
            // 標準は FPS/4 くらい（1/4秒）
            inputController.SetRepeatStartCount(GL.FPS / 4);

            return true;
        }

        /// <summary>
        /// 終了処理
        /// </summary>
        /// <returns>true:正常 false:異常</returns>
        private bool Final()
        {
            // とりあえず何もしない

            return true;
        }

        /// <summary>
        /// 描画処理
        /// </summary>
        private void Draw()
        {
            // 背景
            background.Draw();

            // フィールド（○×）描画
            fieldData.Draw();

            // メッセージ
            message.Draw();

            // プレイヤー操作ブロック
            playerPiece.Draw();
            
            return;
        }

        /// <summary>
        /// 演算処理
        /// </summary>
        private void Perform()
        {
            // ピースの生成処理
            {
                // ピースが無いとき
                if (playerPiece.IsCreated() == false)
                {
                    // 次のピース登場までの待ち時間
                    if (playerPieceStatus.FallPerform())
                    {
                        // ピースを生成する
                        playerPiece.Create(nextPiece.Next());

                        // 衝突していた場合はゲームオーバー
                        if (playerPiece.IsCollision(fieldData))
                        {
                            // ステータスを“終了”に
                            playerStatus = PlayerStatus.End;

                            // メッセージを設定
                            message.Set("GAME OVER");
                        }
                    }
                }
            }

            // キーボードの読み取り
            inputController.Read();

            // 状態により分岐
            switch (playerStatus)
            {
                case PlayerStatus.InputWait: // 入力待ち
                    // 入力待ち処理
                    InputWaitPerform();
                    break;

                case PlayerStatus.End: // 終了
                    // 何もしない
                    return;
            }

            // スライド処理。ピースが生成されているときのみ実施。
            if (playerPiece.IsCreated())
            {
                // スライド処理をして１段以上落下する状態なら落下処理
                if (playerPieceStatus.SlidePerform())
                {
                    // 落下できなくなるまでループ
                    while (playerPieceStatus.IsSlide())
                    {
                        // プレイヤー操作ブロックが着地した場合は、フィールドにセット
                        if (playerPiece.IsDrop(fieldData) == false)
                        {
                            // フィールドにセット
                            playerPiece.SetPiece(fieldData);
                            // ブロックをリセット
                            playerPiece.Remove();
                            // スライド処理をリセット
                            playerPieceStatus.SlideReset();
                            // 抜ける
                            break;
                        }

                        // 一段落下させる
                        playerPiece.Move(0, 1, fieldData);

                        // スライド処理に１段落下を通知
                        playerPieceStatus.SlideDrop();
                    }
                }
            }

            // 削除処理
            fieldData.Shave();

            // 勝敗の判定
            DecideWinner();
        }

        /// <summary>
        /// オブジェクトの更新処理
        /// </summary>
        private void Update()
        {

            
        }

        /// <summary>
        /// 勝敗を決める
        /// </summary>
        private void DecideWinner()
        {
            // 終了している場合は何もしない
            if (playerStatus == PlayerStatus.End) return;
        }

        /// <summary>
        /// プレイヤー入力待ち処理
        /// </summary>
        private void InputWaitPerform()
        {
            // プレイヤー操作ブロックが存在しないときは処理しない
            if (playerPiece.IsCreated() == false) return;

            // Zキー：左回転（反時計回り）
            if (inputController.isKeyPush(DX.KEY_INPUT_Z))
            {
                playerPiece.Rotate(3, fieldData);
            }
            // Xキー：右回転（時計回り）
            if (inputController.isKeyPush(DX.KEY_INPUT_X))
            {
                playerPiece.Rotate(1, fieldData);
            }
            // 左キー：左移動
            if (inputController.isKeyPush(DX.KEY_INPUT_LEFT) ||
                inputController.isKeyRepeat(DX.KEY_INPUT_LEFT))
            {
                playerPiece.Move(-1, 0, fieldData);
            }
            // 右キー：右移動
            if (inputController.isKeyPush(DX.KEY_INPUT_RIGHT) ||
                inputController.isKeyRepeat(DX.KEY_INPUT_RIGHT))
            {
                playerPiece.Move(1, 0, fieldData);
            }
            // 下キー：ドロップ
            if (inputController.isKeyPress(DX.KEY_INPUT_DOWN))
            {
                // スライドキーを押下したことを通知する
                playerPieceStatus.SlideKeyDown();
            }
        }
    }
}
