﻿using DxLibDLL;
using System;

namespace Tetris
{
    /// <summary>
    /// キーボード押下状態制御
    /// </summary>
    class InputController
    {
        // キーボード押下状態を保存するバッファ
        private byte[,] keyStateBuf = new byte[2, 256];

        // バッファのフレーム（どちら側が有効か）
        private int keyStateBufNo = 0;

        // キーボードリピートを制御するためのカウンタ値
        private int[] keyRepeatCount = new int[256];

        // キーボードリピートを開始するまでのカウンタ値
        private int keyRepeatStartCount = 0;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public InputController()
        {
            // バッファの初期化
            keyStateBuf.Initialize();

            // フレームの初期化
            keyStateBufNo = 0;

            // リピートカウンタの初期化
            keyRepeatCount.Initialize();
        }

        /// <summary>
        /// キーボードリピート状態（押し続けて一定期間が過ぎたかどうか）の
        /// カウンタ値をセットする。
        /// このカウンタ値以上キーを押し続けたらリピート状態と見なす
        /// </summary>
        /// <param name="count"></param>
        public void SetRepeatStartCount(int count)
        {
            keyRepeatStartCount = count;
        }

        /// <summary>
        /// キーボード押下状態を読み込む。
        /// </summary>
        public void Read()
        {
            // フレームを切り替える
            keyStateBufNo ^= 1;

            // 今のフレームにキーボード押下状態を読み込む
            DX.GetHitKeyStateAll(out keyStateBuf[keyStateBufNo, 0]);

            // 押下状態のキーのリピートカウンタを増加させる
            for (int i = 0; i < 256; i++)
            {
                // キーが押下状態であれば、カウンタを増加する
                if (keyStateBuf[keyStateBufNo, i] == 1)
                {
                    keyRepeatCount[i]++;

                    // リピート開始していたら、それ以上は増やさない
                    if (keyRepeatCount[i] > keyRepeatStartCount)
                    {
                        keyRepeatCount[i] = keyRepeatStartCount;
                    }
                }
                // キーが押下状態になければ、カウンタをクリアする
                else
                {
                    keyRepeatCount[i] = 0;
                }
            }
        }

        /// <summary>
        /// キーボード押下状態の判定（押し下げ状態かどうか）
        /// </summary>
        /// <param name="keycode">判定するキーコード</param>
        /// <returns>true:押されている false:放されている</returns>
        public bool isKeyPress(int keycode)
        {
            // 今のフレーム状態で判定する
            return (keyStateBuf[keyStateBufNo, keycode] == 1);
        }

        /// <summary>
        /// キーボード押下状態の判定（今、押されたかどうか）
        /// </summary>
        /// <param name="keycode">判定するキーコード</param>
        /// <returns>true:押された false:押されていない or 押しっぱなし</returns>
        public bool isKeyPush(int keycode)
        {
            // 今のフレーム：押されている AND 前のフレーム：放されている のとき true
            return (
                (keyStateBuf[keyStateBufNo, keycode] == 1) &&
                (keyStateBuf[keyStateBufNo ^ 1, keycode] == 0));
        }

        /// <summary>
        /// キーボードリピート状態の判定（押し続けて一定期間が過ぎたかどうか）
        /// </summary>
        /// <param name="keycode">判定するキーコード</param>
        /// <returns>true:リピート状態 false:リピート状態ではない</returns>
        public bool isKeyRepeat(int keycode)
        {
            // 今のフレーム：押されている AND リピート状態 のとき true
            return (
                (keyStateBuf[keyStateBufNo, keycode] == 1) &&
                (keyRepeatCount[keycode] >= keyRepeatStartCount));
        }
    }
}
