﻿using System;

namespace Tetris
{
    /// <summary>
    /// ブロックの位置を表すクラス
    /// 基本的には基点座標からの相対位置を表す
    /// </summary>
    public class Pos
    {
        // 座標
        private int x;
        private int y;

        // アクセサ（取得のみ）
        public int X { get { return this.x; } }
        public int Y { get { return this.y; } }

        /// <summary>
        /// X軸座標,Y軸座標を初期化時に与える
        /// </summary>
        /// <param name="x">X軸座標</param>
        /// <param name="y">Y軸座標</param>
        public Pos(int x, int y) { this.x = x; this.y = y; }

        /// <summary>
        /// 引数なしコンストラクタの禁止
        /// </summary>
        private Pos() { }
    }

    /// <summary>
    /// ブロック４個の組み合わせ（いわゆる「テトリミノ」）を表すクラス
    /// </summary>
    public class Piece
    {
        // 座標（基点は0,0で、他に3個の座標を持つ）
        private Pos[] posList = new Pos[3];

        // 回転法則（0=通常の回転法則 1=無回転=O型）
        private int rotateRule = 0;

        /// <summary>
        /// ３点の座標と回転法則を初期化時に与える。
        /// 回転法則は省略時は0=通常の回転法則となる。
        /// （もし３点より多くても少なくても、普通に動く…はず。テストはしてない）
        /// </summary>
        /// <param name="posList">座標（３点の配列）</param>
        /// <param name="r">回転法則（0=通常の回転法則 1=無回転=O型）</param>
        public Piece(Pos[] posList, int r = 0) { this.posList = posList; this.rotateRule = r; }

        /// <summary>
        /// 引数なしコンストラクタの禁止
        /// </summary>
        private Piece() { }

        /// <summary>
        /// ブロックの位置を取得する。基点を0,0として４個のブロックの相対座標を返す。
        /// （基点0,0＋初期化時点で与えられた座標を返す。
        /// 　初期化時点で３点未満・４点以上与えられていたら、その分も返す）
        /// </summary>
        /// <param name="r">ブロックの向き。0が標準。1～3が右90,180,270度回転（時計回り）</param>
        /// <returns>座標（４個分）</returns>
        public Pos[] GetPos(int r = 0)
        {
            // 戻り値となるPos
            Pos[] p = new Pos[posList.Length + 1];

            // 回転法則が1=無回転の場合は回転させない
            if (rotateRule == 1)
            {
                // 回転を0（正位置にする）
                r = 0;
            }

            // 基点は必ず0,0となる
            p[0] = new Pos(0, 0);

            // 他３個
            for (int i = 0; i < posList.Length; i++)
            {
                // 基準の座標
                int px = posList[i].X;
                int py = posList[i].Y;

                // 回転させる
                for (int j = 0; j < r; j++)
                {
                    // 回転する
                    // X,Y の座標を -Y,X にすると、時計回り90度回転
                    // 1,1 --> -1,1 --> -1.-1 --> 1,-1
                    int wx = px;
                    int wy = py;
                    px = -wy;
                    py = wx;
                }

                // 回転後のPosをセット
                p[1 + i] = new Pos(px, py);
            }

            return p;
        }
    }
    /// <summary>
    /// ピース生成クラス
    /// </summary>
    class PieceFactory
    {
        // ピースの形状データ定義
        // 0,0を回転軸として、他３個のブロックの位置を表す
        private Pos[,] pieceDefine = new Pos[7,3] {
            //
            // I型
            //
            //  □□□□    □□■□    □□□□    □■□□
            //  ■■★■    □□■□    □□□□    □★□□
            //  □□□□    □□★□    ■★■■    □■□□
            //  □□□□    □□■□    □□□□    □■□□
            //
            { new Pos( -2,  0), new Pos( -1,  0), new Pos(  1,  0) },
            //
            // O型
            //
            //  □□□□
            //  □★■□
            //  □■■□
            //  □□□□
            //
            { new Pos(  1,  0), new Pos(  1,  1), new Pos(  0,  1) },
            //
            // J型
            //
            //  ■□□     □■■     □□□     □■□
            //  ■★■     □★□     ■★■     □★□
            //  □□□     □■□     □□■     ■■□
            //
            { new Pos( -1, -1), new Pos( -1,  0), new Pos(  1,  0) },
            //
            // L型
            //
            //  □□□     ■■□     □□■     □■□
            //  ■★■     □★□     ■★■     □★□
            //  ■□□     □■□     □□□     □■■
            //
            { new Pos( -1,  1), new Pos( -1,  0), new Pos(  1,  0) },
            //
            // S型
            //
            //  □□□     ■□□     □■■     □■□
            //  □★■     ■★□     ■★□     □★■
            //  ■■□     □■□     □□□     □□■
            //
            { new Pos(  1,  0), new Pos(  0,  1), new Pos( -1,  1) },
            //
            // Z型
            //
            //  □□□     □■□     ■■□     □□■
            //  ■★□     ■★□     □★■     □★■
            //  □■■     ■□□     □□□     □■□
            //
            { new Pos( -1,  0), new Pos(  0,  1), new Pos(  1,  1) },
            //
            // T型
            //
            //  □■□     □■□     □□□     □■□
            //  ■★■     □★■     ■★■     ■★□
            //  □□□     □■□     □■□     □■□
            //
            { new Pos( -1,  0), new Pos(  0, -1), new Pos(  1,  0) },
        };

        /// <summary>
        /// ピースを生成して戻す
        /// </summary>
        /// <param name="kind">ピースの種別</param>
        /// <returns>生成したピース</returns>
        public Piece CreatePiece(int kind)
        {
            // 範囲外のときはnullを返す
            if (kind < 1 || kind > 7) return null;

            Pos[] pos = new Pos[3];
            for (int i = 0; i < 3; i++)
            {
                pos[i] = new Pos(pieceDefine[kind - 1, i].X, pieceDefine[kind - 1, i].Y);
            }
            return new Piece(pos, (kind == 2 ? 1 : 0));
        }
    }
}
