﻿using System;
using System.Diagnostics;

namespace Tetris
{
    /// <summary>
    /// プレイヤーの操作するピース（落下中の「テトリミノ」）
    /// </summary>
    class PlayerPiece
    {
        // ピースの基準座標
        int x;
        int y;
        // ピースの種別
        int kind;
        // ピースの向き
        int roll;

        // Piece
        Piece piece;

        /// <summary>
        /// ピースが生成されているか
        /// </summary>
        /// <returns>true:生成されている false:生成されていない</returns>
        public bool IsCreated()
        {
            return (piece != null);
        }

        /// <summary>
        /// ピースの生成
        /// </summary>
        /// <param name="kind">生成するピースの種別</param>
        public void Create(int kind)
        {
            this.x = 4; // 中央
            this.y = 0; // 最上段
            this.kind = kind; // 種別
            this.roll = 0; // 方向
            this.piece = new PieceFactory().CreatePiece(kind); // ピース
        }

        /// <summary>
        /// ピースの削除
        /// </summary>
        public void Remove()
        {
            this.kind = 0;
            this.piece = null;
        }

        /// <summary>
        /// ピースの移動。移動先の相対座標を x,y で指定する。
        /// 移動できるかどうか判断し、問題なければ移動する。
        /// </summary>
        /// <param name="x">移動先のX座標（相対位置）</param>
        /// <param name="y">移動先のY座標（相対位置）</param>
        /// <param name="field">移動可能・不可能の判定をするためのField</param>
        /// <returns>true:移動OK false:移動NG</returns>
        public bool Move(int x, int y, Field field)
        {
            Debug.Assert(piece != null);    // デバッグ処理

            // pieceが生成されていないときはfalseを返す（バグでしかありえない）
            if (piece == null) return false;

            // ブロック座標を取得
            Pos[] pos = piece.GetPos(roll);

            // 全ての座標について移動可不可の判定
            foreach (Pos p in pos)
            {
                // 既にブロックがある場合は移動不可(false)を返す
                if (field.Get(this.x + x + p.X, this.y + y + p.Y) != 0)
                {
                    return false;
                }
            }

            // OKなら移動
            this.x += x;
            this.y += y;

            return true;
        }

        /// <summary>
        /// ピースの回転。回転する方向 r を指定する。
        /// 回転できるか判定し、問題なければ回転する。
        /// </summary>
        /// <param name="r">回転の方向</param>
        /// <param name="field">回転可能・不可能の判定をするためのField</param>
        /// <returns>true:回転OK false:回転NG</returns>
        public bool Rotate(int r, Field field)
        {
            Debug.Assert(piece != null);    // デバッグ用

            // pieceが生成されていないときはfalseを返す（バグでしかありえない）
            if (piece == null) return false;

            // 今の状態を保存
            int old_roll = roll;
            
            // 暫定で回転
            roll = (roll + r) % 4;

            // 回転出来るかどうかチェックして、問題なければ回転する
            // 壁蹴り、繰り上がり、T-Spinなどの回転法則はとりあえず後回し

            // ブロック座標を取得
            Pos[] pos = piece.GetPos(roll);

            // 全ての座標についてチェック
            foreach (Pos p in pos)
            {
                // 既にブロックがある場合は回転不可(false)を返す
                if (field.Get(this.x + p.X, this.y + p.Y) != 0)
                {
                    roll = old_roll; // 元に戻してからキャンセル
                    return false;
                }
            }

            // OK
            return true;
        }

        /// <summary>
        /// ブロックが落下可能かどうかの判断。
        /// 一つ下に移動出来るかどうかチェックして、移動できれば“落下可能”と判断する
        /// “落下不可能”＝“接地した”ということ
        /// </summary>
        /// <param name="field">移動可能・不可能の判定をするためのField</param>
        /// <returns>true:落下可能 false:落下不可能</returns>
        public bool IsDrop(Field field)
        {
            Debug.Assert(piece != null);    // デバッグ用

            // pieceが生成されていないときはfalseを返す（バグでしかありえない）
            if (piece == null) return false;

            // ブロック座標を取得
            Pos[] pos = piece.GetPos(roll);

            // 全ての座標についてチェック
            foreach (Pos p in pos)
            {
                // 既にブロックがある場合は移動不可(false)を返す
                if (field.Get(this.x + p.X, this.y + 1 + p.Y) != 0)
                {
                    return false;
                }
            }

            // OK
            return true;
        }

        /// <summary>
        /// ブロックが壁に衝突しているかどうかの判断。
        /// </summary>
        /// <param name="field">衝突の判定をするためのField</param>
        /// <returns>true:衝突している false:衝突していない</returns>
        public bool IsCollision(Field field)
        {
            Debug.Assert(piece != null);    // デバッグ用

            // pieceが生成されていないときはfalseを返す（バグでしかありえない）
            if (piece == null) return false;

            // ブロック座標を取得
            Pos[] pos = piece.GetPos(roll);

            // 全ての座標についてチェック
            foreach (Pos p in pos)
            {
                // 既にブロックがある場合は衝突(true)を返す
                if (field.Get(this.x + p.X, this.y + p.Y) != 0)
                {
                    return true;
                }
            }

            // 衝突していない
            return false;
        }

        /// <summary>
        /// ピースをフィールドにセットする。
        /// セットする先に既にブロックがあった場合は上書きする。
        /// </summary>
        /// <param name="field">セットする先のField</param>
        /// <returns>常にtrue</returns>
        public bool SetPiece(Field field)
        {
            Debug.Assert(piece != null);    // デバッグ用

            // pieceが生成されていないときはfalseを返す（バグでしかありえない）
            if (piece == null) return false;

            // ブロック座標を取得
            Pos[] pos = piece.GetPos(roll);

            // 全ての座標のブロックをセット
            foreach (Pos p in pos)
            {
                field.Set(this.x + p.X, this.y + p.Y, this.kind);
            }

            // OK
            return true;
        }

        /// <summary>
        /// 描画処理
        /// </summary>
        public void Draw()
        {
            // pieceが生成されていないときはfalseを返す（ピース生成前の状態）
            if (piece == null) return;

            // 描画用
            BlockDrawer block = BlockDrawer.Instance;

            // ブロック座標を取得
            Pos[] pos = piece.GetPos(roll);

            // 全ての座標のブロックを描画
            foreach (Pos p in pos)
            {
                // 基準座標からの相対座標を指定する
                block.Draw(x + p.X, y + p.Y, kind);
            }
        }
    }
}
