﻿using System;

namespace Tetris
{
    /// <summary>
    /// プレイヤー操作ブロックを制御するための状態など
    /// （正直、どこに持たせるべきか悩ましい。PlayerPieceの管理下に置くべきかも？）
    /// </summary>
    class PlayerPieceStatus
    {
        // 落下カウンタ
        // カウンタ値=1000で1ライン分の落下
        // レベルが上昇すると一気に2～20ライン落下することもある
        private int slideCount = 0;

        // 強制落下フラグ
        // 要は↓キー押下時に強制落下させるためのもの
        private bool slideKeyDown = false;

        // 次のブロック登場までの待ち時間
        // カウンタ値が0になったら次のブロック登場
        private int fallWaitCount = 0;

        /// <summary>
        /// カウンタ値のリセット
        /// </summary>
        public void ResetCount()
        {
            slideCount = 0;
            slideKeyDown = false;
            fallWaitCount = 0;
        }

        /// <summary>
        /// 次のブロック登場OKか？
        /// </summary>
        /// <returns>true:登場してOK false:登場してはダメ</returns>
        public bool IsFall()
        {
            return (fallWaitCount <= 0);
        }

        /// <summary>
        /// 次のブロック登場までの待ち時間の演算処理と判定処理。
        /// 待ち時間があれば減少させる。
        /// 待ち時間がなくなればtrueを返す。カウンタはリセットする。
        /// </summary>
        /// <returns>true:待ち時間なし（次のブロック登場OK） false:待ち時間あり（次のブロック登場ダメ）</returns>
        public bool FallPerform()
        {
            fallWaitCount--;
            if (fallWaitCount <= 0)
            {
                fallWaitCount = GL.FPS / 2; // 固定値で 1/2秒 調整可能なようにすると良い
                return true;
            }
            return false;
        }

        /// <summary>
        /// 落下OKか？
        /// </summary>
        /// <returns>true:1段以上落下する状態 false:1段も落下しない状態</returns>
        public bool IsSlide()
        {
            return (slideCount >= 1000);
        }

        /// <summary>
        /// スライドカウンタの演算処理と判定処理。
        /// カウンタを増加させ、1段以上落下する状態ならtrueを返す。
        /// </summary>
        /// <returns>true:1段以上落下する状態 false:1段も落下しない状態</returns>
        public bool SlidePerform()
        {
            // とりあえず1秒で1段落下するようにカウンタ値を増加させる
            // 調整可能なようにすると良い
            slideCount += (1000 / GL.FPS);

            // もし↓キー押下などによりスライドの予約があったとき
            // １段下がる状態でなければ、１段だけ落下させる。
            // すでに１段下がる状態のときは、何もしない。
            // これはレベルが上がったときに一気に2～20段落下するのに対応するため。
            // 下を押しても押さなくても変わらない、という状態を作り出すため。
            if (slideKeyDown)
            {
                if (slideCount < 1000) slideCount = 1000;
                slideKeyDown = false;
            }

            // 条件の判定を行う
            if (IsSlide())
            {
                return true;
            }
            else
            {
                return false;
            }

            // この辺りの関数の作りはどうもイマイチな感じがする。
            // GameMainの処理と機能分担がきちんと出来ていない。
        }

        /// <summary>
        /// スライドカウンタのリセット
        /// </summary>
        public void SlideReset()
        {
            slideCount = 0;
        }

        /// <summary>
        /// スライドカウンタを1段分減少する
        /// </summary>
        public void SlideDrop()
        {
            slideCount -= 1000;
        }

        /// <summary>
        /// 強制落下フラグをONにする（↓キー押下時に呼ぶ）
        /// </summary>
        public void SlideKeyDown()
        {
            slideKeyDown = true;
        }
    }
}
